  procedure filesys;

  const
  numsects = 43;

  type
   filedrivenamedef = array[1..numsects] of string[25];

  const
    mostfiles = 500;
    soh = 1;
    eot = 4;
    ack = 6;
    nak = $15;
    can = $18;
    C = $43;
    drivecap = 5000; {Kbyte capacity of files drive}
    ksize = 1; {minimum increment of file size in Kbytes}


 type
    filerec = record
               title: name;
               description : person;
               submit: integer;
               date: name;
               size: integer;
               accesses: integer;
               ASCII: boolean;
               section: byte;
               Xpublic: boolean;
             end;
    channel = array[0..127] of byte;

  var
    filefile: file of filerec;
    filetab: array[0..mostfiles] of filerec;
    filebuff: array [0..16] of channel;
    datafile: file;
    chksum: byte;
    CRC: integer;
    crcmode: boolean;
    enddir: integer;
    comch: char;
    inch : byte;
    filedrive    : longname;
    filesection  :byte;


procedure initfile; forward;
procedure closefile; forward;

procedure listFilesections;
  var
    loopvar : integer;
    temp    : line;
  begin
    if cts then begin
      clearsc;
      lineout('File Sections:' + cr + lf);
      for loopvar := 1 to numsects do begin
        lineout(sect1[loopvar]);
      end;
    end;
  end;

  procedure calcCRC(data:byte);
    var
      carry: boolean;
      i: byte;
    begin
      chksum := lo(chksum + data);
      for i := 0 to 7 do begin
        carry := (crc and $8000) <> 0;
        crc := crc shl 1;
        if (data and $80) <> 0 then crc := crc or $0001;
        if carry then crc := crc xor $1021;
        data := lo(data shl 1);
      end;
    end;

  procedure sendcalc(ch : byte);
    begin
      xmit(ch);
      calcCRC(ch);
    end;

  procedure acknak( time: integer);
   begin
      inch := 0;
      repeat
        if timedin(time*1000) then  inch := inbyte;
      until (inch in [ack, nak, can, C]) or  not cts;
    end;

  function acknakout(ch : byte): boolean;
    var  times, loops: integer;
    begin
      times := 0;
      repeat
        loops := 0;
        xmit(ch);
        while (loops < 10) and not timedin(1000) do loops := loops + 1;
        times := times + 1;
      until inready or (times > 9) or not cts;
      acknakout := inready and cts;
    end;

  procedure download(var successful: boolean);

    var
       loop: byte;
      blocknum, period, tries: integer;
      done: boolean;
      temp: line;
      RecordsRead : integer;

    begin
      FILEMODE := 0;  {0=Read only 2=ReadWrite}
      reset(datafile);
      str(filesize(datafile):4, temp);
      blockread(datafile, filebuff[0], 1,RecordsRead);
      if RecordsRead = 0 then lineout('ERROR: This file is empty!')
      else
      begin
        done := false;
        tries := 0;
        blocknum := 1;
        crcmode := false;
        purgeline(ComBase);
        lineout('Ready for XMODEM transfer:');
        lineout('File open:' + temp + ' records;');
        Statusline(4,'Total Records ='+temp);
        lineout('To cancel: type CTL-X until you return to command prompt.');

        repeat
          acknak(60);
          if inch = 0 then inch := can;
          if inch = C then begin
            crcmode := true;
            StatusLine(3,'CRC mode requested');
          end;
          if inch = ack then begin
            if RecordsRead = 0 then done := true else begin  { ***Turbo 4 fix }
              StatusLine1(2,'Sent #'+IString(blocknum));
              blockread(datafile, filebuff[0], 1,RecordsRead);
              blocknum := blocknum + 1;
              tries := 0;
            end;
          end
          else tries := tries + 1;
          if (inch <> can) and cts and not done then begin
            xmit(soh);
            xmit(lo(blocknum));
            xmit(255-lo(blocknum));
            chksum := 0;
            crc := 0;
            for loop := 0 to 127 do sendcalc(filebuff[0][loop]);
            calcCRC(0);
            calcCRC(0);
            if crcmode then begin xmit(hi(crc)); xmit(lo(crc)); end
              else xmit(chksum);
          end;
          if tries = 5 then crcmode := not crcmode;
        until (inch = can) or done or (tries= 10) or not cts;
        successful := done;
        tries := 0;
        if successful and cts then repeat
          xmit(eot);
          acknak(5);
          tries := tries + 1;
        until (inch=ack) or (tries > 10) or not cts;
        if cts and (inch <> can) and not successful then xmit(can);
        if successful then StatusLine(3,'Transfer completed successfully');
      end;   { this file is empty! }
       close(datafile);
    end;

  function recchar(var error: boolean): byte;
    var temp: byte;
    begin
      temp := 0;
      if not cts then error := true;
      if not error then begin
        if not timedin(1000) then  error := true
        else begin
          temp := inbyte;
          calcCRC(temp);
          recchar := temp;
        end;
      end;
    end;

  procedure clearline;
  var junk: byte;
  begin
    while timedin(1000) do junk := inbyte;
  end;

{$I-}
  procedure upload(var successful: boolean);
    var
      blocknum, tries, byteloc : integer;
      comp, locblock, crc2     : integer;
      fatal, error, done       : boolean;
      opening, inch, locrc     : byte;
      hicrc, csum2, mode       : byte;
    begin
      rewrite(datafile);
      fatal := ioresult > 0;
      tries := 0;
      done := false;
      opening := 0;
      locblock := 1;
      lineout('Beginning XMODEM protocol upload:');
      lineout('To cancel: type CTRL-X until you return to command prompt.');
      if crcmode then mode := C else mode := nak;
      if cts and not fatal then fatal := not acknakout(mode);
      while cts and not (done or fatal) do begin
        tries := tries + 1;
        error := false;
        opening := recchar(error);
        if opening = can then fatal := true;
        if opening = eot then  done := true;
        if (opening <> eot) and (opening <> soh) and not fatal
          then error := true;
        if cts and not (error or fatal or done) then begin
          blocknum := recchar(error);
          comp := recchar(error);
          if lo(comp + blocknum + opening) <> 0 then
          begin
           StatusLine(3,'error-lo(comp+blocknum+opening)');
           error := true;
          end;
          byteloc := 0;
          crc := 0;
          chksum := 0;
          while (byteloc < 128) and not (error or fatal) do begin
            filebuff[0][byteloc] := recchar(error);
            byteloc := byteloc + 1;
          end;
          if cts and not (error or fatal) then begin
            calcCRC(0);
            calcCRC(0);
            crc2 := crc;
            csum2 := chksum;
            hicrc := recchar(error);
            if crcmode then begin
              locrc := recchar(error);
              if (lo(crc2) <> locrc) or (hi(crc2) <> hicrc) then
              begin
                 StatusLine(3,IString(tries)+' CRC error (Hi,Lo) ');
                 error := true;
              end;
            end else if csum2 <> hicrc then
             begin
               StatusLine(3,IString(tries)+' Checksum error (csum2,hicrc)');
               error := true;
            end;
            if (lo(locblock) <> blocknum)
              and (lo(locblock) <> lo(blocknum+1))
              and not error
              then fatal := true;
            if (lo(locblock) = blocknum) and not (error or fatal) then begin
              blockwrite(datafile, filebuff[0], 1);
              StatusLine1(2,' Received #'+IString(blocknum) +'  '+IString(locblock));
              if IOresult <> 0 then fatal := true;
              tries := 0;
              locblock := locblock + 1;
            end;
          end;
        end;
       if not (fatal or error) then flush else clearline;
     {   if (fatal or error) then begin clearline; end;   }
        if not done then begin
          if not (error or fatal) then fatal := not acknakout(ack);
          if error and not fatal then begin
            fatal := not acknakout(nak);
            if tries > 6 then crcmode := not crcmode;
          end;
        end;
      end;
      if fatal then begin StatusLine(3,'Fatal detected');  xmit(can); end;
      if done then  begin StatusLine(3,'Finished...'); xmit(ack); end;
      close(datafile);
      successful := (IOresult = 0) and done and not fatal;
      if successful then StatusLine(4,'Transfer was successful');
      if not successful then erase(datafile);
      repeat
          comch := getcap('Press <cr> to continue........ ');
      until (comch = cr) or not cts;
      stringout('Exit Status : ');
      if fatal      then lineout('Fatal : Abort detected');
      if done       then lineout('Done : Finished...');
      if successful then lineout('Transfer was successful');
    end;

  procedure storebuff(var buffernum: byte; var paused, aborted: boolean);
    var loop: byte;
    begin
      loop := 0;
      while (loop < buffernum) and not aborted do begin
        blockwrite(datafile, filebuff[loop], 1);
        if IOresult > 0 then aborted := true;
        loop := loop + 1;
      end;
      if buffernum in [1..16] then filebuff[0] := filebuff[buffernum];
      buffernum := 0;
      repeat xmit(17) until timedin(1000);
      paused := false;
    end;

  procedure textcap(var successful: boolean);
    var
      buffernum, where, loop  : byte;
      cc, cz, paused          : boolean;
      withecho, done, aborted : boolean;
      temp                    : byte;

    begin
      withecho := (getcap('Do you want your text echoed (Y/N) ? ') = 'Y');
      lineout('Beginning text capture: two CTRL-Cs abort, two CTRL-Zs end.');
      cc := false;
      cz := false;
      done := false;
      paused := false;
      buffernum := 0;
      where := 0;
      rewrite(datafile);
      aborted := (IOresult > 0);
      while cts and not (done or aborted) do begin
        if paused then
          if not timedin(1000) then storebuff(buffernum, paused, aborted);
        temp := inbyte;
        if not cts then aborted := true;
        if withecho and outready then xmit(temp);
        if temp = 3 then begin if cc then aborted := true else cc := true; end
          else cc := false;
        if temp = 26 then begin if cz then done := true else cz := true; end
          else cz := false;
        filebuff[buffernum][where] := temp;
        where := where + 1;
        if where > 127 then begin
          where := 0;
          buffernum := buffernum + 1;
        end;
        if buffernum > 14 then begin
          xmit(19);
          paused := true;
        end;
        if buffernum > 16 then aborted := true;
      end;
      if done and cts and not aborted then begin
        buffernum := buffernum + 1;
        storebuff(buffernum, paused, aborted);
      end;
      close(datafile);
      if aborted and (IOresult = 0) then erase(datafile);
    successful := done and (IOresult=0) and not aborted;
    end;
{$I+}

  function File_exists(filename:longname): boolean;
    var found: boolean;
    begin
      filename := filedrive + filename;
      assign(datafile, filename);
      FileMode := 0;  {0=Read only 2=ReadWrite}
      {$I-} reset(datafile) {$I+};
      found := (IOresult = 0);
      if found then close(datafile);
      File_Exists := found;
    end;

  function alpha(filename: name): boolean;
    var strpos: integer;
        okay:   boolean;
        dots:   byte;
    begin
      dots := 0;
      alpha := true;
      if length(filename) > 0 then
        for strpos := 1 to length(filename) do begin
          if filename[strpos] = '.' then dots := dots + 1;
          if not (filename[strpos] in ['.', '-', '_', '0'..'9', 'A'..'Z'])
            then alpha := false;
        end;
      if dots > 1 then alpha := false;
    end;

  function getlegal: name;
    var filename:  name;
         dotpos : integer;
           ext  : boolean;
           fnl  : boolean;
        not_null : boolean;
      alpha_name : boolean;
    begin
      repeat
        if WMCommandWord = 1001 then begin chat; WMCommandWord := 0; end;
        filename := allcaps(getinput('Enter name of file ? ', 12, echo));
        dotpos := pos('.', filename);
        fnl := (length(copy(filename,1,dotpos)) > 1);
        ext := (length(filename) - dotpos) = 3;
        not_null := filename <> '';
        if not_null then
        begin
          alpha_name := alpha(filename);
          if not fnl  then
          begin
            lineout('Error: Filename length too short...');
            lineout('Filename must be of the form : "NAME.EXT"   (without quotes)');
            lineout('NAME should be between 2 to 8 characters...');
          end;
          if not ext then
          begin
             lineout('Error: Extension incorrect....');
             lineout('EXT should be of legal type : TXT, EXE, COM, DOC  etc...');
          end;
          if not alpha_name then
          begin
            lineout('Error: Non-alpha characters in name...');
            lineout('Acceptable characters are [A...Z] [0...9] [-] [_]'+cr);
          end;
        end;
      until ((fnl and ext and alpha_name) or (not cts) or (not not_null));
      getlegal := filename;
      StatusLine(1,filename);
   end;

   function newname(tabloc: integer): name;
    var filename: name;
    begin
      newname := filetab[tabloc].title;
      stringout('New name? ');
      filename := getlegal;
      if (filename <> '') then begin
        if not File_Exists(filename) then begin
          assign(datafile, filedrive + filetab[tabloc].title);
          rename(datafile, filedrive + filename);
          newname := filename;
          stringout('File renamed.');
        end
        else lineout('Name in use - cannot rename.');
      end;
    end;

  function file_description: person;
    var tempname : person;
    begin
      repeat
        tempname := getinput('Enter description of file ? ', 27, echo);
        if length(tempname) <= 3 then
                lineout('Need a file description before you can continue: ');
      until ((length(tempname) > 3) or not cts);
      file_description := tempname;
    end;


  function dirpos(filename: name): integer;
    var loopvar: integer;
    begin
      dirpos := 0;
      loopvar := 0;
      repeat
        loopvar := loopvar + 1;
      until (filetab[loopvar].title = filename) or (loopvar >= enddir);
    if filetab[loopvar].title = filename then dirpos := loopvar;
    end;

  function getsect: byte;
    var temp, temp1: integer;
    begin
      if sectsin then repeat
        temp := getint(numsects, 0, 'Which section (0 for no change, ? for list) ? ');
        if temp = -1 then listFilesections else  temp1 := temp;
      until (temp <> -1) or not cts
      else temp1 := 1;
      if temp1 <> 0 then  filedrive := filedrivename[temp1] else
                    begin filedrive := filedrivename[1]; temp1 := 1; end;
      initfile;
      filesection := temp1;
      lineout('FileDrive = ['+filedrive+']  Area='+
                istring(temp1));
      getsect := temp1;
    end;

  procedure addfile(filename: name;descr : person; sectnum: byte; xmodem: boolean);
    begin
      with filetab[enddir + 1] do begin
        title := filename;
        if (length(descr) > 1) then description := descr;
        submit := usernum;
        if clockin then date := timeon;
        assign(datafile, filedrive + filename);
        FileMode := 0;  {0=Read only 2=ReadWrite}
        reset(datafile);
        size := filesize(datafile);
        close(datafile);
        accesses := 0;
        ASCII := not xmodem;
        section := sectnum;
        Xpublic := false;
      end;
    end;

procedure newfile;
    var
      filename   : name;
      successful : boolean;
      sectnum    : byte;
      descr      : person;
      innum      : integer;
      UploadMode : integer;
      xmodem     : boolean;


procedure GetUploadParameters;
begin
    repeat
         lineout('***  File Upload Parameters ***');
         lineout(  '1- File Name   : ' + filename);
         lineout(  '2- Description : ' + descr);
         stringout('3- Mode        : ');
         case UploadMode of
            0: lineout('XMODEM - [C]RC');
            1: lineout('XMODEM - Check[S]um');
            2: lineout('Plain [T]ext File');
         end;
         lineout('0- Continue...');
         lineout(space);
         innum := getint(5, 0, 'Number of parameter to change? ');
         case innum of
           1: begin
                 filename := getlegal;
                 if File_Exists(filename) then
                 begin
                   lineout('File name in use.');
                   filename := '';
                 end;
              end;  { end filename }
           2: descr  := file_description;
           3: UploadMode := getint(2,0,'0=CRC  1=Checksum  2=Text  Choice: ')
         end;
       until (not cts) or (innum = 0);
 end;

    begin
      clearsc;
      filename   := '';
      sectnum    := 0;
      descr      := '';
      UploadMode := 1;
      if enddir >= mostfiles then
      begin
        lineout('No file space available.');
        exit;
      end;
      filename := getlegal;
      if File_Exists(filename) then
      begin
        lineout('File name in use.');
        filename := '';
      end;
      if filename = '' then exit;  { good filename?, then I can continue...}
      descr  := file_description;
      lineout('Directory assignment is ' + FileDrive  + filename);
      GetUploadParameters;
      assign(datafile, filedrive + filename);
      case UploadMode of
        0: begin crcmode := true;  upload(successful); xmodem := true; end;
        1: begin crcmode := false; upload(successful); xmodem := true;end;
        2: begin textcap(successful); xmodem := false; end;
      end;
      if successful then addfile(filename, descr, sectnum, xmodem);
      clearline;
      if successful then enddir := enddir + 1  else lineout('*** Fatal Error ***');
   end;

  function legaltab(prompt: line): integer;
    var filename: name;
        tabloc:   integer;
    begin
      tabloc := 0;
      clearsc;
      stringout(prompt);
      filename := getlegal;
      if filename <> '' then begin
        tabloc := dirpos(filename);
        if tabloc <> 0 then
          if not (filetab[tabloc].Xpublic or (access > reg)) then tabloc := 0;
        if tabloc <> 0 then assign(datafile, filedrive + filename)
          else if filename <> '' then lineout('No such file available.');
      end;
      legaltab := tabloc;
    end;

  procedure transmitfile;
    var
      successful: boolean;
      tabloc: integer;
    begin
      tabloc := legaltab('Download: ');
      if tabloc > 0 then begin
        download(successful);
        if successful then with filetab[tabloc] do
          accesses := accesses + 1
        else lineout('Transfer failed.');
      end;
    end;

  procedure textdump;
    var
      tabloc : integer;
      libname: longname;
    begin
      tabloc := legaltab('ASCII text dump: ');
      lineout(space);
      if tabloc > 0 then with filetab[tabloc] do begin
        libname := title;
        typefile(filedrive + libname, true);
        if not cancelled then accesses := accesses + 1;
      end;
    end;

procedure showspace;
var temp : name;
begin
  Str(DiskFree(3),verylongname);
  if cts then
    lineout(cr + lnfd + 'Disk space available : ' +verylongname +'  bytes.');
end;

  procedure dir(sectnum : byte);
    var loop, spaces : byte;
        howbig, sectmin : integer;
        any  : boolean;
        temp : line;
    begin
      any := false;
      sectmin := ksize shl 3;
      lineout(space);
      if sectsin then lineout('Section ' + sect1[sectnum] + ':');
      if enddir > 0 then for loop := 1 to enddir do with filetab[loop] do begin
        howbig := (size + sectmin - 1) div sectmin;
        if cts and (Xpublic or (access = sysop) or (submit = usernum))
         { *** and (sectnum = section)} then begin
          str(howbig:4, temp);
          for spaces := length(title) to 13 do temp := ' ' + temp;
          stringout(title + temp + 'K');
          if clockin then stringout('   ' + date);
          lineout(' ' + description);
          if (access = sysop) or (submit = usernum) then begin
            str(accesses:4, temp);
            stringout('                      D/L='
                + temp + ' ' + getname(submit));
            if not Xpublic then lineout('*Private*') else lineout(' ');
          end;
          any := true;
        end;
      end;
      if cts and not any then lineout('No files found.');
    end;

  procedure directory;
     begin
      stringout('Directory: ');
    {  writeln(filesection);  *** added to trace file section select }
      dir(filesection);
      showspace;
    end;


  procedure killfile;
    var loop, tabloc: integer;
    begin
      tabloc := legaltab('Delete: ');
      if tabloc > 0 then begin
        lineout('Filename is removed from BBS directory, not hard drive...');
        if enddir > tabloc then for loop := tabloc + 1 to enddir do
          filetab[loop - 1] := filetab[loop];
        enddir := enddir - 1;
      end;
    end;

  procedure jiffykillfile(tabloc : integer);
    var loop: integer;
    begin
      if tabloc > 0 then
      begin
        lineout('Filename '+filetab[tabloc].title+' is killed...');
        if enddir > tabloc then for loop := tabloc + 1 to enddir do
          filetab[loop - 1] := filetab[loop];
        enddir := enddir - 1;
      end;
    end;

  procedure editheader(filename: name);
    var tabloc     : integer;
        innum      : integer;
        sectstring : name;
    begin
      if filename <> '' then  inbuffer := filename;
      tabloc := legaltab('main\file\EditHeader: ');
      if tabloc > 0 then with filetab[tabloc] do begin
        repeat
          lineout(space);
          lineout('1- File Name   : ' + title);
          lineout('2- Description : ' + description);
          lineout('3- From User   : ' + getname(submit));
          if access = sysop then lineout('4- Public?     : ' + yn[Xpublic]);
          lineout(space);
          innum := getint(4, 0, 'Number of parameter to change? ');
          case innum of
            1: title          := newname(tabloc);
            2: description    := file_description;
            3: repeat submit  := getid('Name of submitter? ') until (submit <> 0) or not cts;
            4: Xpublic := not Xpublic;
          end;
        until (innum = 0) or not cts;
        assign(datafile, filedrive + title);
        FileMode := 0;  {0=Read only 2=ReadWrite}
        reset(datafile);
        size := filesize(datafile);
        close(datafile);
      end else lineout('File not in directory.');
    end;

 procedure installfile(filename : name);
    var descr    : person;
        sectnum  : byte;
    begin
      if enddir < mostfiles then begin
        filename := getlegal;
        if filename <> '' then begin
          if File_Exists(filename) and (dirpos(filename) = 0) then begin
            descr := '';
            { if (sectnum = 0) then exit; }
            addfile(filename,descr, sectnum, true);
            enddir := enddir + 1;
            lineout('File installed.');
          end;
          editheader(filename);
        end;
      end;
    end;


  procedure initfile;
    var
      loopvar: integer;
      temp: name;
    begin
      if filesopen then closefile;
      lineout('Initializing file system...');
      loopvar := 0;
      verylongname := filedrive + 'FILES.BBS';
      assign(filefile, verylongname);
      FileMode := 0;  {0=Read only 2=ReadWrite}
      {$I-} reset(filefile) {$I+};
      if IOresult = 0 then begin
        while not eof(filefile) do begin
          loopvar := loopvar + 1;
          read(filefile, filetab[loopvar]);
        end;
        close(filefile);
      end;
      enddir := loopvar;
      filesopen := true;
    end;

  procedure closefile;
    var loopvar: integer;
    begin
      if filesopen then
      begin
        rewrite(filefile);
        if enddir > 0 then
          for loopvar := 1 to enddir do write(filefile, filetab[loopvar]);
      close(filefile);
      end;
      filesopen := false;
    end;

procedure jiffyloopkill;
var loopvar : integer;
begin
   for loopvar := 1 to enddir do
   begin
     write('[',filedrive,filetab[loopvar].title,']  ');
     filetab[loopvar].section := filesection;
     if not File_Exists(filetab[loopvar].title) then
     begin
       jiffykillfile(loopvar);
       exit;
     end
     else writeln;
   end;
 end;

procedure jiffy;
var  done : integer;
begin
   repeat
     done := enddir;
     jiffyloopkill;
   until done = enddir;
end;


  begin
  {  initfile;  }
    filedrive := '';
    listfilesections;
    getsect;
    StatusLine(1,'FileMenu...');
    Statusline1(2,'...');
    StatusLine1(3,'...');
    StatusLine1(4,'...');
    if ((not expert) and (inbuffer = '')) then outfile(filemenu);
    repeat
      lineout(space);
      if WMCommandWord = 1001 then begin chat; WMCommandWord := 0; end;
      comch := getcap('main\FILE MENU (?-help) cmd: ');
      if ((comch in['U']) and (access<file_reg)) then
      begin
        lineout('You must be a level 3 access this function........');
        lineout('Please send the SYSOP a note requesting level 3...');
      end;

      case comch of
       'A' : GetSect;
       'D' : directory;
       'S' : begin
               transmitfile;
             end;
       'T' : begin
              textdump;
             end;
       'H' : outfile(filehelp);
       'G' : inbuffer := '-;G';
       '?' : outfile(filemenu);
       'U' : if access>reg then newfile;
       'K' : if access = sysop then killfile;
       'I' : if access = sysop then installfile('');
       'E' : if access = sysop then editheader('');
       'J' : if access = sysop then jiffy;
      end;
    until (comch = '-') or (comch = cr) or not cts;
    if cts then lineout('Closing file system...');
    closefile;
  end;
