

{=======================================================================

                         ASYNC

               Turbo Pascal Asynch Functions

========================================================================}

Type tComPort =  (Com1, Com2, Com3, Com4, Com5 );
     tBaud = (b110, b150, b300, b600, b1200, b2400, b4800, b9600);
     tParity = (pSpace, pOdd, pMark, pEven, pNone);
     tDatabits = (d5, d6, d7, d8);
     tStopbits = (s1, s2);
  
                         {  Define address adders for the various
                             Async card registers.                    }
Const RBR = $00;         { xF8   Receive Buffer Register             }
      THR = $00;         { xF8   Transmitter Holding Register        }
      IER = $01;         { xF9   Interrupt Enable Register           }
      IIR = $02;         { xFA   Interrupt Identification Register   }
      LCR = $03;         { xFB   Line Control Register               }
      MCR = $04;         { xFC   Modem Control Register              }
      LSR = $05;         { xFD   Line Status Register                }
      MSR = $06;         { xFE   Modem Status Register               }
      DLL = $00;         { xF8   Divisor Latch Least Significant     }
      DLM = $01;         { xF9   Divisor Latch Most  Significant     }
      imlist1: array[Com1..Com5] of integer = ($EF, $F7, $FB, $FB, $FB);{ ASynch Interrupt Masks }
      imlist2: array[Com1..Com5] of integer = ($FF, $FF, $FB, $F7, $EF);{ ASynch Interrupt Masks }
      ivlist:  array[Com1..Com5] of integer = ($000C, $000B, $0072, $0073, $0074);{  ASynch hardware interrupt addresses }
      PICCMD1 = $20;           {  8259 Priority Interrupt Controller  }
      PICMSK1 = $21;           {  8259 Priority Interrupt Controller  }
      PICCMD2 = $A0;           {  8259 Priority Interrupt Controller  }
      PICMSK2 = $A1;           {  8259 Priority Interrupt Controller  }
      ComBaseAddr : array[Com1..Com5] of integer = ($3F8,$2F8,$3E8,$3E0, $2F0);
      baudcode: array[b110..b9600] of integer = ($417, $300, $180, $C0, $60, $30, $18, $0C);
      paritycode: array[pSpace..pNone] of byte  = ($38, $08, $28, $18, $00);
      databitscode: array[d5..d8] of byte = ($00, $01, $02, $03);
      stopbitscode: array[s1..s2] of byte = ($00, $04);
      baudrate     :  array[b110..b9600] of integer
            = (110,150,300,600,1200,2400,4800,9600);

      MaxRingBuff = $7FFF;


var       datardy       : boolean;
          ringbuf       : pchar;
          readptr       : integer;
          writptr       : integer;
          ComBase_TXRDY : boolean;
          ComBase_CTS   : boolean;

          datardy_packet   : boolean;
          ringbuf_packet   : pchar;
          readptr_packet   : integer;
          writptr_packet   : integer;
          PacketBase_TXRDY : boolean;
          PacketBase_CTS   : boolean;

          datardy_msys   : boolean;
          ringbuf_msys   : pchar;
          readptr_msys   : integer;
          writptr_msys   : integer;
          MSYSBase_TXRDY : boolean;
          MSYSBAse_CTS   : boolean;


          ComSaveVec : array[Com1..Com5] of Pointer;

Const       ComBase    : integer = $3E8;
            PacketBase : integer = $3E0;
            MsysBase   : integer = $2F0;

{-------This routine gets control upon an Asynch Interrupt---------------}
{F+}
Procedure Port_Interrupt; Interrupt;
 var ComBase_IIR, PacketBase_IIR, MSYSBase_IIR : integer;

Begin
  inline($FA);
  
  { Get the cause(es) of the interrupt - High Priority  }
  ComBase_IIR    := PORT[ComBase +    IIR];
  PacketBase_IIR := PORT[PacketBase + IIR];
  MSYSBase_IIR   := PORT[MSYSBase +   IIR];

  case ComBase_IIR of
        $02 : ComBase_TXRDY    := true;
        $04 : begin
                 ringbuf[writptr] := CHR(PORT[ComBase]);
                 writptr := (writptr + 1) and MaxRingBuff;
                 datardy := true;
              end;
        $00 : begin
                ComBase_MSR := PORT[ComBase +  MSR];
                ComBase_CTS := (ComBase_MSR and $80) = $80;
              end;
   end;

   case PacketBase_IIR of
         $02 : PacketBase_TXRDY := true;
         $04 : begin
                  ringbuf_packet[writptr_packet] := CHR(PORT[PacketBase]);
                  writptr_packet := (writptr_packet + 1) and MaxRingBuff;
                  datardy_packet := true;
               end;
         $00 : begin
                  PacketBase_MSR := PORT[PacketBase + MSR];
                  PacketBase_CTS := (PacketBase_MSR and $80) = $80;
               end;
   end;

   case MSYSBase_IIR of
         $02 : MSYSBase_TXRDY   := true;
         $04 : begin
                 ringbuf_MSYS[writptr_MSYS] := CHR(PORT[MSYSBase]);
                 writptr_MSYS := (writptr_MSYS + 1) and MaxRingBuff;
                 datardy_MSYS := true;
               end;
         $00 : begin
                 MSYSBase_MSR := PORT[MSYSBase +   MSR];
                 MSYSBase_CTS :=  (MSYSBase_MSR and $80) = $80;
               end; 
   end;

   { Service Received data - Low Priority }
   ComBase_LSR    := PORT[Combase    + LSR];
   PacketBase_LSR := PORT[PacketBase + LSR];
   MSYSBase_LSR   := PORT[MSYSBase   + LSR];

  PORT[PICCMD1] := $20;                 {  Send End Of Interrupt to 8259       }
  PORT[PICCMD2] := $20;
  inline($FB);
End;
{F-}

function hex(number : integer): string;
var    i : integer;
       k : integer;
     res : string;
const ch : array[0..15] of char =
 ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F');
begin
   res[0] := #2;
   for i := 1 to 2 do
   begin
     res[3-i] := ch[number and $0f];
     number := number shr 4;
   end;
   hex := res;
end;

function hex4(number : integer): string;
var    i : integer;
       k : integer;
     res : string;
const ch : array[0..15] of char =
 ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F');
begin
   res[0] := #4;
   for i := 1 to 4 do
   begin
     res[5-i] := ch[number and $0f];
     number := number shr 4;
   end;
   hex4 := res;
end;


procedure PortStatus(combaseX : integer);
begin
 SerialPort_Update(1);
 SerialPort_Update(2);
 SerialPort_Update(3);
 StatusLIne(0,'$'+hex4(combaseX)+
         ' RBR=$'+hex(port[combaseX+RBR])+
         ' MCR=$'+hex(port[combaseX+MCR])+
         ' MSR=$'+hex(port[combaseX+MSR])+
         ' IIR=$'+hex(port[combaseX+IIR])+
         ' IER=$'+hex(port[combaseX+IER])+
         ' LSR=$'+hex(port[combaseX+LSR]));
end;


procedure setbaudrate(ComPort: tComport;
                      Baud:tBaud;
                      Parity: tParity;
                      Databits: tDatabits;
                      Stopbits: tStopbits);
var LCRreg   : byte;
    temp     : integer;
    ComBaseX : integer;
begin
  ComBaseX := ComBaseAddr[ComPort];
  LCRreg := $80;                                    {  Divisor Latch Access         }
  LCRreg := LCRreg or paritycode[Parity];           {  Setup Parity                 }
  LCRreg := LCRreg or databitscode[Databits];       {  Setup # data bits            }
  LCRreg := LCRreg or stopbitscode[Stopbits];       {  Setup # stop bits            }
  PORT[IER+ComBasex] := $00;                        {  Disable Data Avail interrupt }
  PORT[MCR+ComBasex] := $00;                        {  Restore Com interrupt vector }
  PORT[LCR+ComBaseX] := Port[LCR + ComBase] or  $80;{  Reset DLAB                   }
  PORT[DLM+ComBaseX] := Hi(baudcode[Baud]);         {  Set Baud rate                }
  PORT[DLL+ComBaseX] := Lo(baudcode[Baud]);         {  Set Baud rate                }
  PORT[LCR+ComBaseX] := LCRreg and $7F;             {  Reset DLAB                   }
  PORT[IER+ComBaseX] := $0F;                        {  Enable some interrupts       }
  PORT[MCR+ComBaseX] := $0B;                        {  Set RTS, DTR, OUT2           }
  StatusLine(4,'Port $'+hex4(ComBaseX)+' set to '+IString(Baudrate[Baud]));
  delay(200);
  PortStatus(CombaseX);
end;


{-------------------- Open COM1 or COM2, a la Basic -----------------}
Procedure OpenCom(ComPort: tComPort;
                  Baud: tBaud;
                  Parity: tParity;
                  Databits: tDatabits;
                  Stopbits: tStopbits);

var ComBaseX : integer;

Begin
  inline($FA);
  ComBaseX := ComBaseAddr[ComPort];           {  Select Input Port        }
    StatusLine(1,'Opening ComPort $'+hex4(CombaseX));
  if (port[ComBaseX+IER] <> 0) then
  begin
    StatusLine(2,'Warning!!!');
    StatusLine(3,'Program Halted because ComPort $'+hex4(CombaseX));
    StatusLine(4,'is already Opened.......');
    PORT[IER+ComBasex] := $00;              {  Disable Data Avail interrupt }
    PORT[MCR+ComBasex] := $00;              {  Restore Com interrupt vector }
  {  Halt;  }
  end;
  PORT[MCR+ComBaseX] := $00;              {  Restore Com interrupt vector }
  GetIntVec(ivlist[ComPort], ComSaveVec[ComPort]);
  SetIntVec(ivlist[Comport],@PORT_Interrupt);
  readptr := 0;                              {  Init buffer pointers     }
  writptr := 0;                              {  Init buffer pointers     }
  datardy := false;
  readptr_packet := 0;                       {  Init buffer pointers     }
  writptr_packet := 0;                       {  Init buffer pointers     }
  datardy_packet := false;
  readptr_MSYS := 0;                       {  Init buffer pointers     }
  writptr_MSYS := 0;                       {  Init buffer pointers     }
  datardy_MSYS := false;
  ComBase_TXRDY    := false;
  PacketBase_TXRDY := false;
  MSYSBase_TXRDY   := false;
  PORT[PICMSK1] := PORT[PICMSK1] and imlist1[Comport];{  Enable ASynch Int }
  PORT[PICMSK2] := PORT[PICMSK2] and imlist2[Comport];{  Enable ASynch Int}
  setbaudrate(ComPort,Baud,Parity,Databits,Stopbits);
  PORT[IER+ComBaseX] := $0F;                  {  Enable some interrupts   }
  PORT[MCR+ComBaseX] := $0B;                  {  Set RTS, DTR, OUT2       }
  inline($FB);
End;


{---------------- Close any initialized COM -------------------------}
Procedure RestoreV(ComPort : tComPort);
var ComBasex : integer;
Begin
  ComBasex := CombaseAddr[ComPort];
  PORT[PICMSK1] := PORT[PICMSK1] or ($FF - imlist1[Comport]); { Disable interrupt }
  PORT[PICMSK2] := PORT[PICMSK2] or ($FF - imlist2[Comport]); { Disable interrupt }
  PORT[IER+ComBasex] := $00;              {  Disable Data Avail interrupt }
  PORT[MCR+ComBasex] := $00;              {  Restore Com interrupt vector }
  SetIntVec(ivlist[ComPort], ComSaveVec[ComPort]);
End;

Procedure RestoreVector;
begin
  RestoreV(Com3);
  RestoreV(Com4);
  RestoreV(Com5);
  PortStatus(ComBase);
  PortStatus(PacketBase);
  PortStatus(MSYSBase);
end;

function get_rcv_char: char;
begin
  get_rcv_char := ringbuf[readptr];
  readptr := (readptr + 1) and MaxRingBuff;
  inline($FA);
  if (readptr = writptr) then datardy := false;
  inline($FB);
end;

Function  recvchar: char;
Begin
  If datardy then recvchar := char(ord(get_rcv_char) and $7f);
end;



{=======================  PACKET ===================}

function get_rcv_char_packet: char;
begin
  get_rcv_char_packet := ringbuf_packet[readptr_packet];
  readptr_packet := (readptr_packet + 1) and MaxRingBuff;
  inline($FA);
  if (readptr_packet = writptr_packet) then datardy_packet := false;
  inline($FB);
end;

Function  recvchar_packet: char;
Begin
  If datardy_packet then recvchar_packet := char(ord(get_rcv_char_packet) and $7f);
end;

procedure xmitchar_packet(ch : char);
Var LSRreg: byte;
  begin
    { Repeat LSRreg := (PORT[LSR+PacketBase] and $20) Until LSRreg <> 0; }
     Repeat until PacketBase_TXRDY or (not cts);
     PacketBase_TXRDY := false;
     PORT[THR+PacketBase] := ORD(ch);
  end;

procedure stringout_packet(message : line);
var charpos : integer;
begin
for charpos := 1 to length(message) do xmitchar_packet(message[charpos]);
end;

function cts_packet: boolean;
  begin
   { PacketBase_MSR := PORT[PacketBase + MSR]; }
    cts_packet := PacketBase_CTS;  { (PacketBase_MSR and $80) = $80; }
 end;

{=======================  END  PACKET  ===================}
{=======================  MSYS ===================}

function get_rcv_char_MSYS: char;
begin
  get_rcv_char_MSYS := ringbuf_MSYS[readptr_MSYS];
  readptr_MSYS := (readptr_MSYS + 1) and MaxRingBuff;
  inline($FA);
  if (readptr_MSYS = writptr_MSYS) then datardy_MSYS := false;
  inline($FB);
end;

Function  recvchar_MSYS: char;
Begin
  If datardy_MSYS then recvchar_MSYS := char(ord(get_rcv_char_MSYS) and $7f);
end;

procedure xmitchar_MSYS(ch : char);
Var LSRreg: byte;
  begin
    { Repeat LSRreg := (PORT[MSYSBase+LSR] and $20) Until LSRreg <> 0; }
     Repeat until MSYSBase_TXRDY or (not cts);
     MSYSBase_TXRDY := false;
     PORT[THR+MSYSBase] := ORD(ch);
  end;

procedure stringout_MSYS(message : line);
var charpos : integer;
begin
for charpos := 1 to length(message) do xmitchar_MSYS(message[charpos]);
end;

function cts_MSYS: boolean;
  begin
    {MSYSBase_MSR := PORT[MSYSBase + MSR]; }
    cts_MSYS := MSYSBase_CTS;      { (MSYSBase_MSR and $02) = $00;}
 end;

procedure dropRTS_MSYS;
{ Lowers RS-232 RTS line - used to inhibit auto-answer
   and to cause modem to hang up }
  begin
      PORT[MCR+MSYSBase] := $09;    {  Reset RTS, DTR, OUT2       }
  end;

procedure raiseRTS_MSYS;
{ Lowers RS-232 RTS line - used to inhibit auto-answer
   and to cause modem to hang up }
  begin
      PORT[MCR+MSYSBase] := $0B;    {  Reset RTS, DTR, OUT2       }
  end;



{=======================  END  MSYS  ===================}

procedure purgeline;
var temp : char;
begin
 repeat
   if datardy then temp := recvchar;
   delay(10);
 until not datardy;
end;

function inready: boolean;
begin
  inready := keypressed or datardy;
end;

function outready: boolean;
begin
 outready := ComBase_TXRDY;  {( (PORT[LSR+ComBase] and $20) <> 0); }
end;

function cts_temp: boolean;
  begin
   { ComBase_MSR := PORT[Combase + MSR]; }
    cts_temp := ComBase_CTS;   { ((ComBase_MSR and $80) = $80) or local; }
  end;


{This function returns true if a carrier tone is present on the modem
 and is frequently checked to see if the caller is still present.
 It always returns "true" in the local mode.  The delays also provide for
 some of the spike noise that is present on telephone lines.......}
var cts_count : integer;
function cts: boolean;
  var cts_temp_var : boolean;
  begin
    cts_temp_var := cts_temp;
    if cts_temp_var then cts_count := 0 else cts_count := cts_count + 1;
    if ((not cts_temp_var) and (cts_count < 5)) then
    begin
       delay(30);
       cts_temp_var := cts_temp;
    end;
    cts := cts_temp_var;
  end;

procedure xmitchar(ch : char);
Var LSRreg: byte;
  Begin
  if (not local and cts) then
  begin
     {Repeat LSRreg := (PORT[LSR+ComBase] and $20) Until LSRreg <> 0; }
     Repeat until ComBase_TXRDY or (not cts);
     ComBase_TXRDY := false;
     PORT[THR+ComBase] := ORD(ch);
  end;
  WindowsMessageHandler;
end;

 procedure xmit(x:byte);
    begin
      xmitchar(chr(x));
    end;

  function inbyte: byte;
    var temp: char;
    begin
      repeat until inready or not cts;
      if keypressed then temp := readkey else
          if datardy then temp := get_rcv_char;
      inbyte := ord(temp);
    end;


{==========================  Modem Commands =========================}
procedure x_xmitchar(ch : char);
Var LSRreg: byte;
  begin
    { Repeat LSRreg := (PORT[LSR+ComBase] and $20) Until LSRreg <> 0; }
     Repeat until ComBase_TXRDY or (not cts);
     ComBase_TXRDY := false;
     PORT[THR+ComBase] := ORD(ch);
  end;

procedure x_stringout(buffer : line);
var loop: integer;  temp : char;
begin
  for loop := 1 to length(buffer) do
   begin
      temp := buffer[loop];
      if temp <> cr then write(temp);
      x_xmitchar(temp);
   end;
end;

procedure x_quit;
begin
  restorevector;
  writeln(cr+'Problem encountered with accessing the MODEM!  Program Halted!');
  repeat  delay(4000); messagebeep(0); until keypressed;
  halt;
end;

procedure modemdelay(xxxx : integer);
  begin
    TimeOutTimer := false;
    SetTimer(CrtWindow,0,xxxx,nil);
    while ((not TimeOutTimer) and (not datardy)) do WindowsMessageHandler;
  end;

procedure modem_cmd(buffer : line);
var    i : integer;
    temp : char;
begin
   i := 0;
   repeat
     x_stringout(buffer);
     modemdelay(5000);
     if not datardy then i := i + 1;
   until (i = 2) or datardy;

   if i=2 then x_quit;  { serious problem with modem detected }
   {nosound;  }
   if datardy then
   begin
     write(' = [');
     repeat
        temp := recvchar;
        case temp of
          cr   : begin end;         { _write('<cr>');   }
          lnfd : begin end;         { _write('<lnfd>'); }
          else begin write(temp); modemdelay(50); end;
        end;
        delay(50);
     until not datardy;
     writeln(']');
   end;
end;

procedure clearmodem;
var i : integer;
begin
   modem_cmd(cr + 'AT' + cr);
end;

procedure initmodem;        (* Modem Dependent *)
var init_string : line;
  begin
    clearmodem;
    init_string := cr + 'AT X4 Q0 S2=126 S0=1 V0 E0 M1 &C1 &D2 &W' + cr;
    modem_cmd(cr + 'AT Z &F' + cr);
    modem_cmd(init_string);
    modem_cmd(init_string);

   {
                     DEFAULT CODES SENT TO MODEM
      &C1  - DCD carrier detect         X4     - extended command set
      Q0   - no result codes            S2=126 - tild '~' escape char
      S0=1 - rings to answer            V0     - numeric response code
      E0   - no echo in command mode    M1     - monitor speaker till connect
      &D3  - DTR disconnect & reset     &W     - writes config to memory
      &F   - Reset to factory defaults
   }
  end;


procedure dropRTS;
{ Lowers RS-232 RTS line - used to inhibit auto-answer
   and to cause modem to hang up }
  begin
      PORT[MCR+ComBase] := $00;    {  Reset RTS, DTR, OUT2       }
  end;

procedure raiseRTS;
(* Raises RTS line to enable auto-answer *)
  begin
    PORT[MCR+ComBase] := $0B;      {  Set RTS, DTR, OUT2       }
  end;

procedure setlocal;
{Sets local flag true and inhibits modem auto-answer}
  begin
    modem_cmd(cr + 'AT H1 M0' + cr);
    local := true;
    ScreenON;
  end;

procedure clearlocal;
{Clears local flag and allows modem auto-answer}
  begin
    modem_cmd(cr + 'AT H0 M1' + cr);
    local := false;
    ScreenOFF;
  end;


procedure dispcaller;
{Displays caller's data on the diaglog panel}
  begin
    StatusLine(1,'A='+IString(ord(access))+' '+caller);
    StatusLine(2,calls+'  '+tele);
    StatusLIne(3,add1);
    StatusLine(4,add2);
    UserParameter_Update;
  end;

procedure hangup_seq;
begin
  StatusLine(3,'Hangup sequence....');
  dropRTS;
  repeat until not cts;
  delay(50);
  raiseRTS;
end;


procedure hangup;
  begin
    if local then clearlocal else
    begin
       if cts then lineout('--- PLEASE DISCONNECT ---' + cr + lf);
       hangup_seq;
    end;
  end;




